#include "stage1.h"

/*------------------------------------------------------------------------*/
static void
search_p(curr_poly_t *curr, curr_rat_factor_t *curr_factors,
		knapsack_data_t *knap, bounds_t *bounds, 
		stage1_stat_t *stats, unsigned int p0_max)
{
	int i;
	unsigned int r0;
	unsigned int p0 = 1;

	for (i = 0; i < curr_factors->p0_list_len; i++) {
		curr_factors->last_p0 = p0;
		p0 = curr_factors->p0_list[i].p;
		r0 = curr_factors->p0_list[i].r;
		if (p0 > p0_max)
			break;

		stats->n_p++;
		mpz_mul_ui(curr->gmp_prod, curr->gmp_prod_base, p0);

		if (p0 == 1) {
			init_knapsack_exact(curr, knap, curr_factors,
					bounds->a3_max);
		}
		else {
			init_knapsack_exact_p0(curr, knap, curr_factors,
					p0, r0, bounds->a3_max);
		}
		knapsack_exact(curr, knap, bounds, stats);
	}
}

/*------------------------------------------------------------------------*/
static void
search_a5(curr_poly_t *curr, rat_factor_t *rat_factors,
		curr_rat_factor_t *curr_factors,
		knapsack_data_t *knap,
		bounds_t *bounds, stage1_stat_t *stats)
{
	int i, j;
	double p_size, dp0_max, dbl_a5, dbl_N;
	unsigned int p0_max;
	unsigned int shift = 0;
	mpz_t gmp_tmp;
	uint32 npr_in_p;
	uint32 npr_excess;
	uint32 *p_ind;
	curr_rat_prime_t *primes;

	dbl_N = mpz_get_d(curr->gmp_N);
	dbl_a5 = mpz_get_d(curr->gmp_a5);
	stage1_bounds_update(bounds, dbl_N, dbl_a5);
	mpz_init(gmp_tmp);

	while (1) {
		stats->n_a5++;
		shift++;
		if (shift >= 1000) {
			/* update some variables, not very precise 
			   at the beginning of a5-range */
			mpz_set_ui(gmp_tmp, shift);
			mpz_mul_ui(gmp_tmp, gmp_tmp, MULTIPLIER);
			mpz_add(curr->gmp_a5, curr->gmp_a5, gmp_tmp);
			if (mpz_cmp(curr->gmp_a5, bounds->gmp_a5_end) > 0)
				break;
			dbl_a5 = mpz_get_d(curr->gmp_a5);
			stage1_bounds_update(bounds, dbl_N, dbl_a5);
			shift = 0;
		}
		if (!curr_rat_factor_find(rat_factors, 
					curr_factors, bounds->p_size_max)) {
			continue;
		}

		stats->n_pr++;

		/* it actually is much faster to let GMP compute
		   the fifth root from scratch than to use the
		   previous root as a starting point and compute
		   the Newton iteration manually */

		mpz_set_ui(gmp_tmp, shift);
		mpz_mul_ui(gmp_tmp, gmp_tmp, MULTIPLIER);
		mpz_add(curr->gmp_a5, curr->gmp_a5, gmp_tmp);
		shift = 0;

		mpz_fdiv_q(curr->gmp_root, curr->gmp_N, curr->gmp_a5);
		mpz_root(curr->gmp_root, curr->gmp_root, 5);
		mpz_set(curr->gmp_m0, curr->gmp_root);

		if (mpz_cmp(curr->gmp_a5, bounds->gmp_a5_end) > 0)
			break;
		curr_rat_factor_fill(curr_factors, curr->gmp_N,
						curr->gmp_a5);

		p_size = 0;
		dbl_a5 = mpz_get_d(curr->gmp_a5);
		stage1_bounds_update(bounds, dbl_N, dbl_a5);

		p_size = 0;
		p_ind = curr_factors->p_ind;
		primes = curr_factors->primes;
		npr_in_p = curr_factors->npr_in_p;
		npr_excess = curr_factors->npr_total - npr_in_p;
		for (i = 0; i < npr_in_p; i++) {
			p_ind[i] = i;
			p_size += primes[i].p_log;
		}

		curr_rat_factor_fill_aux(curr_factors, 
				curr->gmp_N, curr->gmp_a5,
				exp(bounds->p_size_max - p_size),
				bounds->p0_limit);

		while (1) {
			dp0_max = exp(bounds->p_size_max - p_size);
			if (dp0_max > (double) (bounds->p0_limit))
				p0_max = bounds->p0_limit;
			else
				p0_max = (unsigned int) (rint(dp0_max) - 1);
			if (p0_max < 1)
				p0_max = 1;

			mpz_set_ui(curr->gmp_prod_base, 1);
			for (i = 0; i < npr_in_p; i++) {
				mpz_mul_ui(curr->gmp_prod_base, 
					curr->gmp_prod_base, 
					primes[p_ind[i]].p_pr);
			}

			search_p(curr, curr_factors, knap,
					bounds, stats, p0_max);

			i = npr_in_p - 1;
			while (i >= 0) {
				if (p_size > bounds->p_size_max) {
					i--;
					while ((i >= 0) && (p_ind[i] + 1 ==
							p_ind[i + 1])) {
						i--;
					}
					if (i < 0)
						break;
					p_ind[i]++;
					for (j = i + 1; j < npr_in_p; j++)
						p_ind[j] = p_ind[j - 1] + 1;
					p_size = 0;
					for (j = 0; j < npr_in_p; j++)
						p_size += primes[p_ind[j]].p_log;
					if (p_size <= bounds->p_size_max)
						break;
					continue;
				}
				while ((i >= 0) && (p_ind[i] >= i + npr_excess))
					i--;
				if (i < 0)
					break;
				p_ind[i]++;
				for (j = i + 1; j < npr_in_p; j++)
					p_ind[j] = p_ind[j - 1] + 1;
				p_size = 0;
				for (j = 0; j < npr_in_p; j++)
					p_size += primes[p_ind[j]].p_log;
				if (p_size <= bounds->p_size_max)
					break;
			}
			if (i < 0)
				break;
		}
	}

	mpz_clear(gmp_tmp);
}

/*------------------------------------------------------------------------*/
void
poly_stage1_init(poly_stage1_t *data,
		 stage1_callback_t callback, void *callback_data)
{
	memset(data, 0, sizeof(poly_stage1_t));
	mpz_init_set_ui(data->gmp_N, 0);
	mpz_init_set_ui(data->gmp_a5_begin, 0);
	mpz_init_set_ui(data->gmp_a5_end, 0);
	data->callback = callback;
	data->callback_data = callback_data;
}

/*------------------------------------------------------------------------*/
void
poly_stage1_free(poly_stage1_t *data)
{
	mpz_clear(data->gmp_N);
	mpz_clear(data->gmp_a5_begin);
	mpz_clear(data->gmp_a5_end);
}

/*------------------------------------------------------------------------*/
uint32
poly_stage1_run(poly_stage1_t *data)
{
	uint32 success = 0;
	stage1_stat_t stats;
	bounds_t bounds;
	curr_poly_t currpoly;
	rat_factor_t rat_factors;
	curr_rat_factor_t curr_rat_factors;
	knapsack_data_t knapdata;

	stage1_stat_init(&stats);
	stage1_bounds_init(&bounds, data);
	curr_poly_init(&currpoly, &bounds, data);
	rat_factor_init(&rat_factors, currpoly.gmp_N, currpoly.gmp_a5);
	curr_rat_factor_init(&curr_rat_factors, data->npr_in_p);
	knapsack_data_init(&knapdata, data->npr_in_p);

	search_a5(&currpoly, &rat_factors, &curr_rat_factors, &knapdata,
			&bounds, &stats);

	if (stats.n_survivors)
		success = 1;

	stage1_bounds_free(&bounds);
	curr_poly_free(&currpoly);
	rat_factor_free(&rat_factors);
	curr_rat_factor_free(&curr_rat_factors);
	stage1_stat_free(&stats);
	knapsack_data_free(&knapdata);
	return success;
}
