#ifndef _STAGE2_H_
#define _STAGE2_H_

#include <poly_skew.h>

#ifdef __cplusplus
extern "C" {
#endif

#define NAFF_PRIMES 46

/*-----------------------------------------------------------------------*/
/* data used in the current polynomial */

typedef struct {
	mpz_t gmp_a[6];
	mpz_t gmp_b[6];
	mpz_t gmp_lina[2];
	mpz_t gmp_linb[2];
	mpz_t gmp_help1;
	mpz_t gmp_help2;
	mpz_t gmp_help3;
	mpz_t gmp_help4;
	mpz_t gmp_p;
	mpz_t gmp_d;
	mpz_t gmp_m;
	mpz_t gmp_mb;
} curr_poly_t;

/*-----------------------------------------------------------------------*/
/* data for rating polynomial yield */

typedef struct {
	double bound0;
	double bound1;
	double area;
	unsigned int *prime_list;
	unsigned int prime_bound;
	double alpha_max;
	double alpha_random;
} assess_t;

void assess_init(assess_t *a, double b0, double b1, 
		double area, unsigned int prime_bound);

void assess_free(assess_t *a);

void check(int x, int y, poly_stage2_t *data, double skewness);

int murphy_alpha(double *alpha, int deg, assess_t *assess,
		mpz_t * gmp_coeff, double alpha_targ);

void murphy_alpha_exact(double *alpha, int deg, assess_t *assess,
		    mpz_t * gmp_coeff, unsigned int pb);

void murphy_e_score(double *me, int deg0, double *dbl_coeff0, 
		int deg1, double *dbl_coeff1, double alpha0, 
		double alpha1, double skewness, int nsm, 
		assess_t *assess);

int find_poly_optima(int *deg, double **coeff, 
			double skewness, double *optima);

/*-----------------------------------------------------------------------*/
/* routines for translating polynomials */

void optimize_1(curr_poly_t *c, double *skewness, 
		double *pol_norm, double *alpha_proj);
void optimize_2(curr_poly_t *c, double skewness, 
		double *new_skewness, double *norm_ptr);
void optimize_3(curr_poly_t *c, poly_stage2_t *data, assess_t *assess,
		double skewness, double *new_skewness, 
		double *norm_ptr, double *eptr, double *alphaptr);

/*-----------------------------------------------------------------------*/
/* data for the root sieve */

typedef struct {
	int xmin;
	int xmax;
	int ymin;
	int ymax;
} rs_bound_t;

typedef struct {
	int primepower;		/* p^k */
	int rooti;		/* zero =i mod p^k */
	int J;			/* f(i)+J*(i-m)=0 mod p^k */
	int step;		/* stepwidth p^(k-l) */
	int start;		/* sievearray[start] has zero =i mod p^k */
	int lineinc;		/* increment for change y->y+1 */
	unsigned short v;	/* scaled value */
	double value;		/* log(p)/(p^(k-1)*(p+1)) */
} primelist_t;

typedef struct {
	primelist_t *pl;
	int primelistlen;
	double limit; 
	double st_alpha;
	unsigned short *sievearray;
        unsigned short default_cut;
	int sievelen; 
	int nsubsieves;

	int prep_len; 
	int prep_p_len[NAFF_PRIMES];
	unsigned int *prep_p_begin[NAFF_PRIMES];
	unsigned int *prep_p[NAFF_PRIMES];
} root_sieve_t;

void root_sieve_init(root_sieve_t *rs);
void root_sieve_free(root_sieve_t *rs);
double compute_proj_alpha(curr_poly_t *c);
void root_sieve_run(poly_stage2_t *data, double skewness, 
			double pol_norm, double alpha_proj);

/*-------------------------------------------------------------------------*/

/* data for optimizing a single (a5, p, d) triplet */

typedef struct {
	curr_poly_t curr_poly;
	root_sieve_t root_sieve;
	assess_t assess;
} stage2_curr_data_t;

/*-------------------------------------------------------------------------*/
static INLINE double
ifs(double *coeff, double skewness)
{				/* degree=5 */
	double sq[6], d, s, res;
	int i, j, k;

	for (i = 0; i < 6; i++)
		sq[i] = coeff[i] * coeff[i];
	for (i = 0; i < 4; i++) {
		d = 2 * coeff[i];
		k = i + 1;
		for (j = i + 2; j < 6; j += 2, k++)
			sq[k] += d * coeff[j];
	}
	s = skewness;
	res = 0.;
	res += s * sq[3] / 35.;
	res += sq[2] / s / 35.;
	s *= (skewness * skewness);
	res += s * sq[4] / 27.;
	res += sq[1] / s / 27.;
	s *= (skewness * skewness);
	res += s * sq[5] / 11.;
	res += sq[0] / s / 11.;
	return res;
}


/*-------------------------------------------------------------------------*/
#define COMPUTE_IFS   s2=sc; res=0.; \
  res+=s2*sq[3]/35.; res+=sq[2]/s2/35.;   \
  s2*=(sc*sc);   \
  res+=s2*sq[4]/27.; res+=sq[1]/s2/27.;   \
  s2*=(sc*sc);   \
  res+=s2*sq[5]/11.; res+=sq[0]/s2/11.


static INLINE double
find_best_skewness(double *coeff, double s0)
{				/* degree=5 */
	double sq[6], d, s, sc, s2, res, ds, v;
	int i, j, k;

	for (i = 0; i < 6; i++)
		sq[i] = coeff[i] * coeff[i];
	for (i = 0; i < 4; i++) {
		d = 2 * coeff[i];
		k = i + 1;
		for (j = i + 2; j < 6; j += 2, k++)
			sq[k] += d * coeff[j];
	}
	ds = 10.;
	s = s0;
	sc = s;
	COMPUTE_IFS;
	v = res;
	while (ds > 2.) {
		sc = s + ds;
		COMPUTE_IFS;
		if (res < v) {
			s = sc;
			v = res;
			ds *= 1.1;
			continue;
		}
		if (ds < s) {
			sc = s - ds;
			COMPUTE_IFS;
			if (res < v) {
				s = sc;
				v = res;
				ds *= 1.1;
				continue;
			}
		}
		ds *= 0.9;
	}
	return s;
}

#ifdef __cplusplus
}
#endif

#endif /* !_STAGE2_H_ */
